//
// bop.C
//
// Written by:  Yanting Wang    01/17/2003
//

#include <stdlib.h>

#include "numtools.h"
#include "vector.h"

#include <iostream>
using namespace std;

//
// function w
//
// Calculate bond order parameters q(l) ans w(l).
//
// p[] is the array of bond vectors.
//

double w( int l, const Vector *p, int n, double &q )
{
  if( n < 1 ) return 1.0e-138;

  double cosTheta[n], phi[n];

  for( long i=0; i<n; i++ )
  {
    // Note to convert phi from (-PI/2,PI/2] to (0,2*PI]
    if( Numeric::fzero( p[i].x ) )
    {
      if( Numeric::fzero( p[i].y ) ) phi[i] = 0.0;
      else if( p[i].y > 0 )  phi[i] = Numeric::PI / 2.0;
      else  phi[i] = 3.0 * Numeric::PI / 2.0;
    }
    else 
    {
      phi[i] = atan( p[i].y / p[i].x );
      if( p[i].x < 0.0 )  phi[i] += Numeric::PI;
      else if( p[i].y < 0.0 )  phi[i] += 2.0 * Numeric::PI;
    }

    cosTheta[i] = p[i].z / sqrt(p[i].r2());
  }

  //
  // Average bond order parameters
  //

  complex Q[2*l+1];

  for( int j=0; j<n; j++ )
  {
    Q[ l ] += Numeric::sphHarmonic( l, 0, cosTheta[j], phi[j] );  // m=0
    for( int m=1; m<=l; m++ )
    {
      complex c = Numeric::sphHarmonic( l, m, cosTheta[j], phi[j] );
      Q[ m+l ] += c;
      // m1 = -m
      double sign = (m&1) ? -1.0 : 1.0;
      Q[ -m+l ] += complex( sign * c.r, -sign * c.i );
    }
  }

  for( int m=-l; m<=l; m++ ) Q[ m+l ] /= double(n);

  //
  // Calculate q ( Q[l] )
  //

  q = 0.0;
  for( int m=-l; m<=l; m++ )
  {
    q += Q[ m+l ].a2();
  }
  q = sqrt( 4.0 * Numeric::PI / ( 2.0*l + 1 ) * q );

  //
  // Calculate w
  //

  // wigner 3-j symbols
  int ndim = 2 * l + 1;
  double w3j[ndim];
  double m2min, m2max;   // minimum and maximum values of m2
  int errflag;

  complex numerator;

  for( int m1=-l; m1<=l; m1++ )
  {
    Numeric::ThreeJSymbolM( l, l, l, m1, m2min, m2max, w3j, ndim, errflag );
    if( errflag )
    {
      cerr << " ThreeJSymbolM: Error number " << errflag << endl;
      exit( 1 );
    }

    complex temp;
    for( int m2=int( m2min ); m2<=int( m2max ); m2++ )
    {
      temp += Q[ m2+l ] * Q[ l-m2-m1 ] * w3j[ m2 - int( m2min ) ];
    }
    temp *= Q[ m1+l ];

    numerator += temp;
  }

  double wl = numerator.r;

  double denominator = 0.0;

  for( int m=-l; m<=l; m++ )
  {
    denominator += Q[ m+l ].a2();
  }  

  wl /= sqrt( denominator * denominator * denominator );

  return wl;
}
