//
// calcbop.C
//
// Calculate bond order parameters of each atom for a given configuration.
// 
// Written by: Yanting Wang             01/18/2003
//

#include <iostream>
using namespace std;

#include <libgen.h>
#include <stdlib.h>

#include "cell.h"
#include "vector.h"

double w( int l, const Vector *p, int n, double &q );

Vector *q;   // configuration of particles
int num;    // number of particles
Cell *cell;   // cell index method
int **neighbor;  // indices of neighboring particles
int *nnum;    // number of neighbors

double cutoff;

bool readConf()
{
  if( ! (cin >> num) ) return false;
  char s[1024];
  cin.getline( s, 1024 );
  cin.getline( s, 1024 );

  Vector min, max;

  q = new Vector[num];
  for( int i=0; i<num; i++ )
  {
    double x, y, z;
    cin >> s >> x >> y >> z;

    q[i].set( x, y, z );
    if( i==0 )
    {
      min = q[0];   max = q[0];
    }
    else
    {
      double cx = q[i].x;
      double cy = q[i].y;
      double cz = q[i].z;

      if( min.x > cx ) min.x = cx;   if( max.x < cx ) max.x = cx;
      if( min.y > cy ) min.y = cy;   if( max.y < cy ) max.y = cy;
      if( min.z > cz ) min.z = cz;   if( max.z < cz ) max.z = cz;
    }
  }

  nnum = new int[num];
  neighbor = new int*[ num ];
  for(int i=0; i<num; i++ ) neighbor[i] = new int[ num-1 ];

  // get the largest difference
  Vector d = max - min;
  double diff = d.x;
  if( diff < d.y ) diff = d.y;
  if( diff < d.z ) diff = d.z;

  double size = cutoff + 0.00001;

  cell = new Cell( q, num, (int)ceil(diff/size/2.0)+1, size,
		   (max+min)/2.0 );

  return true;
}

void end()
{
  delete []q;   delete cell;   delete []nnum;
  for( int i=0; i<num; i++ ) delete []( neighbor[i] );
  delete []neighbor;   
}

void genPair()
{
  for( int i=0; i<num; i++ ) nnum[i] = 0;

  cell->update();
  while( cell->gen() )
  {
    int i = cell->p1;   int j = cell->p2;
    if( (q[i] - q[j]).r2() < cutoff*cutoff )
    {
      neighbor[i][nnum[i]++] = j;
      neighbor[j][nnum[j]++] = i;
    }
  }
}

void calcBOP( int average )
{
  if( average )   // average over all bonds
  {
    // get number of pairs
    int bondNum = 0;
    for( int i=0; i<num; i++ ) bondNum += nnum[i];
    bondNum /= 2;

    Vector p[bondNum];

    int k = 0;
    for( int i=0; i<num; i++ )
      for( int j=0; j<nnum[i]; j++ )
      { 
	int m = neighbor[i][j];
	if( i<m ) p[k++] = q[m] - q[i];
      }

    if( k != bondNum )
    {
      cerr << "Error with bond number: " << endl;
      cerr << "k=" << k << " bondNum=" << bondNum << endl;
      end();
      exit(1);
    }

    double w4, q4, w6, q6;
    w4 = w( 4, p, bondNum, q4 );
    w6 = w( 6, p, bondNum, q6 );

    cout << q4 << " " << q6 << " " << w4 << " " << w6 << endl;
  }
  else   // each local structure seperately
  {
    cout << num << endl << endl;

    for( int i=0; i<num; i++ )
    {
      int n = nnum[i] + 1;
      int lq[n];
      lq[0] = i;
      for( int j=1; j<n; j++ ) lq[j] = neighbor[i][j-1]; 

      Vector p[ n*(n-1)/2 ];
      int bondNum = 0;

      for( int j=0; j<n-1; j++ )
      {
	for( int k=j+1; k<n; k++ )
	{
	  Vector v = q[ lq[k] ] - q[ lq[j] ];
	  if( v.r2() < cutoff * cutoff ) p[ bondNum ++ ] = v;
	}
      }

      double w4, q4, w6, q6;
      w4 = w( 4, p, bondNum, q4 );
      w6 = w( 6, p, bondNum, q6 );

      cout << q4 << " " << q6 << " " << w4 << " " << w6 << endl;
    }
  }
}


int main( int argc, char *argv[] )
{
  if( argc < 3 )
  {
    cerr << "Usage: " << basename( argv[0] ) 
	 << " cutoff average?(1:0)" << endl;
    return 1;
  }

  cutoff = atof( argv[1] );

  while( readConf() )
  {
    genPair();
    
    calcBOP( atoi(argv[2]) );  

    end();
  }

  return 0;
}
