//
// cell.C
//
// Written by:  Yanting Wang   Feb. 23, 2002
//
// Implementation of struct cell defined in cell.h
//

#include "cell.h"

#include <stdlib.h>

#include <iostream>
using namespace std;

Cell::Cell( const Vector *p, int num, int hn, double size, const Vector cm )
  : p(p), num( num ), size( size ), n( (hn<=3)?6:2*hn ), cm( cm )
{
  list = new int[num];

  cell = new int **[n];
  for( int i=0; i<n; i++ )
  {
    cell[i] = new int *[n];
    for( int j=0; j<n; j++ )
    {
      cell[i][j] = new int[n];
    }
  }
}

Cell::~Cell()
{
  delete []list;

  for( int i=0; i<n; i++ )
  {
    for( int j=0; j<n; j++ )  delete []cell[i][j];
    delete []cell[i];
  }
  delete []cell;
}

void Cell::update()
{
  update( cm );
}

void Cell::update( const Vector &cm )
{
  //
  // clear arrays
  //
  for( int i=0; i<num; i++ ) list[i] = -1;
  for( int i=0; i<n; i++ )
    for( int j=0; j<n; j++ )
      for( int k=0; k<n; k++ ) cell[i][j][k] = -1;

  // update arrays
  for( int i=0; i<num; i++ )
  {
    Vector v = p[i] - cm;
    int nx = (int)floor( v.x / size ) + n/2;
    int ny = (int)floor( v.y / size ) + n/2;
    int nz = (int)floor( v.z / size ) + n/2;

    if( nx > n-2 || ny > n-2 || nz > n-2 
	|| nx < 1 || ny < 1 || nz < 1 )
    {
      cerr << endl << "Particle out of boundary." << endl;
      cerr << "Please increase the box size." << endl;
      cerr << "Paricle position: " << p[i] << endl;
      cerr << "nx: " << nx << " ny: " << ny << " nz: " << nz << endl;
      exit(1);
    }

    int d = cell[ nx ][ ny ][ nz ];

    if( d == -1 )   // this is the first particle in this cell
    {
      cell[ nx ][ ny ][ nz ] = i;
    }
    else
    {
      while( list[ d ]!= -1 )      // search till the tail
      {
	d = list[ d ];
      }

      list[ d ] = i;
    }
  }

  reset();
}

void Cell::init( bool genCp )
{
  if( genCp ) cp = cell[ xn ][ yn ][ zn ];
  if( cp == -1 ) return;

  status = 0;
  index[ 0 ] = list[ cp ];      // next particle in current cell
  // head of 13 neighbors
  index[ 1 ] = cell[ xn+1 ][ yn-1 ][ zn-1 ];
  index[ 2 ] = cell[ xn+1 ][ yn ][ zn-1 ];
  index[ 3 ] = cell[ xn+1 ][ yn+1 ][ zn-1 ];
  index[ 4 ] = cell[ xn ][ yn+1 ][ zn-1 ];

  index[ 5 ] = cell[ xn+1 ][ yn-1 ][ zn ];
  index[ 6 ] = cell[ xn+1 ][ yn ][ zn ];
  index[ 7 ] = cell[ xn+1 ][ yn+1 ][ zn ];
  index[ 8 ] = cell[ xn ][ yn+1 ][ zn ];

  index[ 9 ] = cell[ xn+1 ][ yn-1 ][ zn+1 ];
  index[ 10 ] = cell[ xn+1 ][ yn ][ zn+1 ];
  index[ 11 ] = cell[ xn+1 ][ yn+1 ][ zn+1 ];
  index[ 12 ] = cell[ xn ][ yn+1 ][ zn+1 ];

  index[ 13 ] = cell[ xn ][ yn ][ zn+1 ];
}

void Cell::reset()
{
  //
  // prepare for pair generating
  //
  out = false;
  xn = 1;   yn = 1;   zn = 1;
  init( true );
}

bool Cell::check()
{
  if( out ) return false;

  //
  // Step to next cell if no more particles in current cell
  //
  if( cp == -1 )
  {
    if( xn < n-2 ) xn++;  // skip the last one for programming simplicity
    else
    {
      if( yn < n-2 ) { yn++; xn = 1; }
      else
      {
	if( zn < n-2 ) { zn++; yn = 1; xn = 1; }
	else { out = true; return false; }    // no more pairs
      }
    }

    init( true );
    return check();
  } 

  //
  // Step to next status
  //
  if( index[ status ] == -1 )    // cp > -1
  {
    if( status < 13 ) status ++;  // step to next status
    else  // status == 13
    {
      cp = list[ cp ];
      init( false ); 
    }

    return check();
  }

  return true;
}

bool Cell::gen()
{
  if( !check() ) return false;

  //
  // Generate a particle pair
  //
  p1 = cp;   p2 = index[ status ];
  index[ status ] = list[ p2 ];

  return true;
}

void Cell::dump()
{
  cout << endl << endl << "***************Cell info***************" << endl;

  cout << "List: " << endl;

  for( int i=0; i<num; i++ ) cout << list[i] << " ";
  cout << endl;

  cout << "Cell: " << endl;

  for( int i=0; i<n; i++ ) 
    for( int j=0; j<n; j++ )
      for( int k=0; k<n; k++ ) 
	if( cell[i][j][k] > -1 ) 
	  cout << i << " " << j << " " << k << " " << cell[i][j][k] << endl;

  cout << "*************End cell info*************" << endl << endl;
}
