//
// cell.h
//
// Cell index method used to keep track of particle neighbors.
// Assume the same sizes for the 3 dimensions of the cells and the central
// point is cm (center of mass) or (0,0,0). This program is for MD without 
// boundary condition.
//
// For programming simplicity, the outside most cells will not be checked. This
// can be ensured by making the region large enough so that no particles can
// go so far.
//
// Written by: Yanting Wang    Feb. 23, 2002
//

#ifndef __CELL_H
#define __CELL_H

#include "vector.h"

class Cell
{
private:
  int *list;      // linked list of particles
  const int num;  // particle number

  int ***cell;    // 3D cell matrix used to store the head of the linked
                   // list of particles
  const int n;     // number of cells in one dimension
  const double size;     // size of one cell

  const Vector *p;   // particle array
  const Vector cm;   // central point 
                     // = ( (maxx+minx)/2.0, (maxy+miny)/2.0, (maxz+minz)/2.0 )

  int xn, yn, zn;  // current cell
  int cp;         // index of current particle
  int status;      // 0: generate pairs in the same cell 
                   // 1: generate pairs with the first neighbor cell
                   // ...
                   // 13: generate pairs with the 13th neighbor cell
  int index[ 14 ];// index of current particles in the current cell
                   // and each of the 13 neighbors
  bool out;        // generating out of pairs

  void init( bool genCp ); // initializing construction for a new focused cell
  bool check();      // check validity and step to next one if invalid

public:
  int p1, p2;     // indices of two interacting particles

  Cell( const Vector *p, int num, int hn, double size, 
	const Vector cm = Vector(0,0,0) );
                     // input half number of cells in 1D
  ~Cell();

  void update();     // update cell and list
  void update( const Vector &cm );
  bool gen();        // generate two interacting particles p1 and p2

  void reset();      // reset to enable gen() from beginning again

  void dump();       // print out list and cell
};


#endif //__CELL_H
